--[[Leopard's In-Save Command Collection
    Install instructions:
        Unzip the save file.

        Put this file next to "control.lua" in the unzipped folder.
        Add "require("leosScripts")" to the start of the "control.lua" file.

        re-zip the folder and check that it has the same folder structure as the original.

        Load the save.
]]


--[[Helper functions]]

local version = 1.1

--[[Resets permssion groups if not defined, or when force is specified (version 1.1)]]
local function ResetPermissionGroups(force)    
    if not global["LeosScript"]["ElevatedPermissionGroups"] or force then
        global["LeosScript"]["ElevatedPermissionGroups"] = {}
        log("Init ElevatedPermissionGroups: (Re)Set PermissionGroupTable.")
    end
end

--[[Check if player has permjssion to use privileged commands: if allowed returns true (version 1.0)]]
local function CheckPermission(player)
    if player.admin then return true end 
    ResetPermissionGroups() --[[Will check if the groups are valid, and reset them if they are not.]]
    if global["LeosScript"]["ElevatedPermissionGroups"][player.permission_group.name] then 
        return true
    end

    --[[When we get there there was nothing that grants permission]]
    return false
end


--[[Edited from https://stackoverflow.com/questions/28664139/lua-split-string-into-words-unless-quoted]]
--[[Separates words from given string but preserves quoted sections (version 1.0)]]
local function SmartWordSeparation(text)
    local spat, epat, buf, quoted = [=[^(['"])]=], [=[(['"])$]=]
    local separatedStrings = {}
    for str in text:gmatch("%S+") do
        local squoted = str:match(spat)
        local equoted = str:match(epat)
        local escaped = str:match([=[(\*)['"]$]=])
        if squoted and not quoted and not equoted then
            buf, quoted = str, squoted
        elseif buf and equoted == quoted and #escaped % 2 == 0 then
            str, buf, quoted = buf .. ' ' .. str, nil, nil
        elseif buf then
            buf = buf .. ' ' .. str
        end
        if not buf then 
            --[[print((str:gsub(spat,""):gsub(epat,"")))]]
            table.insert(separatedStrings, (str:gsub(spat,""):gsub(epat,"")))
        end
    end

    if buf then 
        --[[print("Missing matching quote for ".. buf)]]
        return nil
    end
    --[[log(serpent.block(separatedStrings))]]
    return separatedStrings
end

--[[Commands]]



--[[/CommandPermissions (version 1.3)]]
local CommandPermissionsHelp = "[Admins] Change wich permission groups have access to privileged commands. Use subcommand \"help\" for additional info."
local function CommandPermissions(event)
    local player = game.players[event["player_index"]]
    local parameters = event["parameter"]
    log("CommandPermissions: EventDatas:\n" .. serpent.block(event))
    if not game.player then log("CommandPermissions: event.player was nil?") return end 

    ResetPermissionGroups() --[[Will check if the groups are valid, and reset them if they are not.]]

    if not parameters then        
        player.print("The groups that have permission to use privileged commands are:")
        player.print("\"/promoted\" Admins, use command \"/admins\" to see the Admins.")
        for key, val in pairs(global["LeosScript"]["ElevatedPermissionGroups"]) do
            player.print(key)
        end  
        return
    end      
    
     --[[If parameters ware given:]]
    if not CheckPermission(player) then player.print("You do not have permission to edit the permissions") return end

    local parametersAsWords = SmartWordSeparation(parameters)

    local subCommand = parametersAsWords[1]
    if not subCommand then player.print("Something went wrong parsing your input: a SubCommand was not found!") return end 

    if subCommand == "help" then
        player.print("IMPORTANT: renaming a permissiongroup will make it not work anymore!")
        player.print("empty subCommand: show the current configuration.")
        player.print("subCommand \"clear\": clears all permission groups from the privileged list.")
        player.print("subCommand \"add\" or \"remove\": use the subcommand whitout any arguments to see their helpText.")
        return
    end


    if subCommand == "clear" then 
        global["LeosScript"]["ElevatedPermissionGroups"] = {} 
        player.print("Cleard all permission groups from priviledged permission access. Onely \"/promoted\" Admins will fave access.")
        return 
    end

    if subCommand == "add" then
        if not parametersAsWords[2] then
            player.print("Provide the name(s) \"in quotes\" of the permission group(s) to add.")
            return
        else
            for i=2, #parametersAsWords do
                local groupName = parametersAsWords[i]
                if not groupName then 
                    player.print("Something went wrong iterating your parameters, check /commandpermissions to see what has been changed.") 
                elseif game.permissions.get_group(groupName) then                         
                    global["LeosScript"]["ElevatedPermissionGroups"][groupName] = 1
                    player.print(groupName .. " is now a privileged group.")                        
                else
                    player.print(groupName .. " is not a valid permission group, it will not be added!")
                end
            end
        end 
        return
    end 

    if subCommand == "remove" then
        if not parametersAsWords[2] then
            player.print("Provide the name(s) \"in quotes\" of the permission group(s) to remove.")
            return
        else
            for i=2, #parametersAsWords do
                local groupName = parametersAsWords[i]
                if not groupName then 
                    player.print("Something went wrong iterating your parameters, check /commandpermissions to see what has been changed.") 
                else
                    global["LeosScript"]["ElevatedPermissionGroups"][groupName] = nil
                    player.print(groupName .. " is no longer a privileged group.")                     
                end
            end
        end 
        return
    end 

    --[[SubCommand was not found:]]
    player.print("unknown subCommand: \""  .. subCommand .. "\"  try \"/commandpermissions help\" for help.")    
end
--[[/CommandPermissions end]]

--[[/tp (version 1.0)]]
local tpHelp = "[Admins] Teleports you to the entity you select."
local function tp ()
	if not game.player then log("Command/tp: player was nil?") return end 
	if not game.player.admin then game.player.print("Onely Admins can use this command!") return end 
	
	if game.player.selected then
		local position = game.player.surface.find_non_colliding_position("player", game.player.selected.position, 10, 0.1)
		if not position then game.player.print("Fail: A position tot teleport to was not found.") return end 
		game.player.teleport(position)
	else 
		game.player.print("There was nothing selected!")
	end 	
end 

--[[/tp2me (version 1.0)]]
local tp2meHelp = "[Admins] Teleports the thing you have selected to you."
local function tp2me ()
	if not game.player then log("Command/tp2: player was nil?") return end 
	if not game.player.admin then game.player.print("Onely Admins can use this command!") return end 
	
	if game.player.selected then
		local position = game.player.surface.find_non_colliding_position(game.player.selected.name, game.player.position, 10, 0.1)
		if not position then game.player.print("Fail: A position tot teleport to was not found.") return end 
		game.player.selected.teleport(position) 
	else 
		game.player.print("There was nothing selected!")
	end 
end 

--[[/tp2spawn (version 1.0)]]
local tp2spawnHelp = "[Admins] Teleports you to the exact position of spawn."
local function tp2spawn ()
	if not game.player then log("Command/tp2Spawn: player was nil?") return end 
	if not game.player.admin then game.player.print("Onely Admins can use this command!") return end 

	game.player.teleport(game.player.force.get_spawn_position(game.player.surface))
end 

--[[/nameOfItem (version 1.0)]]
local nameOfItemHelp = "Prints information about the item on your cursor and/or the thing you have selected. [Anyone]"
local function nameOfItem ()
	if not game.player then log("Command/nameOfItem: player was nil?") return end 
	if game.player.cursor_stack.valid_for_read then
		game.player.print("Cursor item name: \"" .. game.player.cursor_stack.name .. "\".")
		game.player.print("Cursor item subgroup: \"" .. game.player.cursor_stack.prototype.subgroup.name .."\", order: \"" .. game.player.cursor_stack.prototype.order .. "\"." )
	else 
		game.player.print("Cursor is empty.")
	end 

	if game.player.selected then
		game.player.print("Selected thing's name: \"" .. game.player.selected.name .. "\".")
		game.player.print("Selected thing's type: \"" .. game.player.selected.type .. "\".")
	else 
		game.player.print("There was nothing selected.")
	end 
end

--[[/walkSpeed (version 1.0)]]

--[[OnGameSpeedChanged (version 1.0)]]
local function OnGameSpeedChanged()
    local walkSpeedMod = global["LeosScript"]["walkSpeedModifier"]
    if not walkSpeedMod then walkSpeedMod = 1 end
    local modifier = ((walkSpeedMod/game.speed) - 1)
    if modifier < -0.9 then modifier = -0.9 end
    game.forces["player"].character_running_speed_modifier = modifier
end

--[[/WalkSpeed (version 1.0)]]
local walkSpeedCommand = "walkspeed"
local walkSpeedHelp = "[Anyone] see the how the walking speed is adjusted. [Admin] use subcommand \"help\" for additional information."
local function WalkSpeed(event)
    local inputString = event["parameter"] 
    local player = game.players[event["player_index"]]
    local walkSpeedMod = global["LeosScript"]["walkSpeedModifier"]
    
    if not player then log("Command/WalkSpeed: player was nil?") return end

    log(serpent.block(event))

    if not inputString then 
        player.print("The walkingSpeed modifier is "..walkSpeedMod.. ". It is applied like so: running_speed_bonus = (modifier/game.speed) - 1).") 
        player.print("The walkingSpeed is adjusted to make the effect of slowing down the game less noticable.")
    return end

    if not player.admin then player.print("You don't have permission to use this command.") return end

    if inputString == "help" then
        player.print("Use \"/walkspeed [new value]\" to edit the walkingSpeed modifier. It is applied like so: running_speed = (modifier/game.speed) - 1).")
        player.print("The default is 1. Valid values range from >0 to <=100. 0.5 means the speed is halved, 2 means the speed is doubled.")
        return
    end
    

    local newValue = tonumber(inputString)
    if not newValue then player.print(inputString .. " could not be onverted to a number!") return end
    if not (newValue > 0) then player.print("The newvalue must be > 0.") return end
    if not (newValue < 100) then player.print("The newvalue must be < 100.") return end

    --[[The newValue is correct.]]
    global["LeosScript"]["walkSpeedModifier"] = newValue
    OnGameSpeedChanged()
    game.print(player.name .. " changed the walkingSpeedModifier to " .. newValue .. ".")
    return
    

    
end

--[[OnSpeedBoundsChanged (version 1.1)]]
local function OnSpeedBoundsChanged(player)
    local lower = global["LeosScript"]["gameSpeedLowerBound"]
    local upper = global["LeosScript"]["gameSpeedUpperBound"]

    if game.speed > upper then
        game.speed = upper
        OnGameSpeedChanged()
        game.print(player.name .. " changed the GameSpeed bounds and the GameSpeed has been changed to " .. game.speed .. " to comply.")
        return
    end
    if game.speed < lower then
        game.speed = lower
        OnGameSpeedChanged()
        game.print(player.name .. " changed the GameSpeed bounds and the GameSpeed has been changed to " .. game.speed .. " to comply.")
        return
    end
end

--[[ResetSpeedBounds (version 1.0)]]
local function ResetSpeedBounds(player)
    global["LeosScript"]["gameSpeedLowerBound"] = 0.05
    global["LeosScript"]["gameSpeedUpperBound"] = 9000    
    OnSpeedBoundsChanged(player)
end

--[[gameSpeedChange (version 1.1)]]
local function gameSpeedChange(player, newSpeed)
    local speedLowerBound = global["LeosScript"]["gameSpeedLowerBound"]
    local speedUpperBound = global["LeosScript"]["gameSpeedUpperBound"]

    if not speedLowerBound or not speedUpperBound then ResetSpeedBounds(player) log("Command/gameSpeed: Speedbounds ware reset because they ware nil!") end
    
	if newSpeed >= 9000 and newSpeed > speedUpperBound then player.print("The Server's Power Level can't handle that kind of Speed!") return end 
	if newSpeed > speedUpperBound then player.print("You may not set the GameSpeed to above " .. speedUpperBound .. ".") return end 
	if newSpeed == 0 then player.print("Don't try to crash the Server! Setting GameSpeed to zero makes no sense!") return end 
	if newSpeed <  0 then player.print("Don't try to crash the Server! Setting GameSpeed to negative makes no sense!") return end 
    if newSpeed < 0.05 then player.print("You can't set the GameSpeed to below 0.05: it is too slow!") return end 
	if newSpeed < speedLowerBound then player.print("You may not set the GameSpeed to below " .. speedLowerBound .. ".") return end 

	log("Command/gameSpeed: We will apply " .. newSpeed .. " as the new gameSpeed.")
	game.speed = newSpeed
    OnGameSpeedChanged()
	game.print(player.name .. " set the game speed to " .. newSpeed)
end


--[[gameSpeedBoundChange (version 1.1)]]
local function gameSpeedBoundChange(player, lower, upper)
    if lower < 0.05  then
        player.print("The lowerBound must be >= 0.05, the game cannot handle speeds below that in multiplayer.")
        return
    end
    if lower > upper then
        player.print("The lowerBound must be less than the upperBound.")
        return
    end

    --[[Nothing was wrong]]
    global["LeosScript"]["gameSpeedLowerBound"] = lower
    global["LeosScript"]["gameSpeedUpperBound"] = upper
    player.print("Bounds are now: >= " .. lower .. " and <=  " .. upper .. ".")
    OnSpeedBoundsChanged(player)
end

--[[/gamespeed (version 1.3)]]
local gameSpeedHelp = "[Anyone] Show the current GameSpeed, [Trusted players] Change the game speed. Use argument \"help\" to see additional information."
local function gameSpeed(event)
    local inputString = event["parameter"] 
    local player = game.players[event["player_index"]]

    local speedLowerBound = global["LeosScript"]["gameSpeedLowerBound"]
    local speedUpperBound = global["LeosScript"]["gameSpeedUpperBound"]

	if not player then log("Command/gameSpeed: player was nil?") return end 
	
	if not inputString then 
		player.print("The game speed is currently set to: " .. game.speed .. " this means " .. (60 * game.speed) .. " UPS.")		
		return 
	end 

	log("Command/gameSpeed: ".. player.name .. " entered \"" .. inputString .. "\" as arguments.")


    if inputString == "help" then
        if not CheckPermission(player) then
            player.print("UPS is Updates Per Second. 60 UPS (gameSpeed 1) is normal. Your FPS is always less than (or equal to) UPS.")
            player.print("The game may be slowed down to allow people on weaker computers to keep playing.")
            return
        end 
        --[[User is privileged]]
        player.print("Use \"/gamespeed [number]\" to change the game speed. Bounds are: >= " .. speedLowerBound .. " and <= " .. speedUpperBound .. "." )
        player.print("Use \"/gamespeed slowest\" to set the gameSpeed to the slowest allowed value.")
        if player.admin then 
            player.print("To change the bounds type \"/gamespeed bounds [lower] [upper]\" and replace the \"[...]\" with the appropriate values.") 
            player.print("To reset the bounds and back to defaults type \"/gamespeed reset\".")
        end
        return
    end

    if not CheckPermission(player) then player.print("You don't have permission to change the game speed.") return end 	

    inputArgs = SmartWordSeparation(inputString)
    --[[Check if the command is to change the speed bounds]]

    local newSpeed = tonumber(inputString)
    
    if inputArgs[1] == "bounds" then
        if not player.admin then player.print("You need to be an admin to change the speed bounds.") return end

        local newLower = tonumber(inputArgs[2])
        local newUpper = tonumber(inputArgs[3])
        if newLower and newUpper then gameSpeedBoundChange(player, newLower, newUpper) return 
        else
            player.print("Arguments could not be converted to a number!")
            return 
        end
    elseif inputArgs[1] == "reset" then
        ResetSpeedBounds(player)
        player.print("Reset speedbounds to defaults.")
        return
    elseif inputArgs[1] == "slowest" then
        gameSpeedChange(player, speedLowerBound)
        return
    elseif inputArgs[1] and not newSpeed then 
        player.print("Unknown argument(s) \"" .. inputString .. "\"")
        return
    end
    

	
    if not newSpeed then player.print(inputString .. " could not be converted into a number!") return end 
	gameSpeedChange(player, newSpeed)
end 
--[[/gamespeed end]]

--[[/Rules (version 1.0)]]

local CommandRulesName = "rules"
local CommandRulesHelp = "Shows the rules on this server."
local function CommandRules(event)    
    local inputString = event["parameter"] 
    local player = game.players[event["player_index"]]
    if not player then log("Command/gameSpeed: player was nil?") return end 

    local ruleList = global["LeosScript"]["RuleList"]
    if not ruleList then 
        global["LeosScript"]["RuleList"] = {} 
        ruleList = {} 
        player.print("There are no rules defined yet.")

        if player.admin then
            player.print("To edit the rules see \"/rules help\"")
        end        
    end

    if not inputString then
        player.print("Rules and Notices:")
        for i = 1, #ruleList do
            player.print("#" .. i .. " " .. ruleList[i])
        end
        return
    end

    if not player.admin then player.print("You do not have permission to supply arguments to this comand.") return end

    if inputString == "help" then
        player.print("Arguments: 1 ==> \"clear\": Clears the list of rules.")
        player.print("Arguments: 2 ==> \"add\": supply a \"quoted rule\" to be added.")
        player.print("Arguments: 2 ==> \"remove\": supply a index of the rule to be removed.")
        player.print("Arguments: 3 ==> \"replace\": supply a index of the rule to be replaced and a \"quoted rule\" for the new value.")
        player.print("Arguments: 3 ==> \"insert\": supply a index of the position for insertion and a \"quoted rule\" to be inserted.")
        return
    end

    if inputString == "clear" then
        global["LeosScript"]["RuleList"] = {} 
        player.print("Rules have been cleared")
        return
    end

    local arguments = SmartWordSeparation(inputString)
    if not arguments[1] then player.print("Something went wrong parsing your input.") return end

    if arguments[1] == "add" then
        if not arguments[2] then player.print("You need to supply a rule to this command") return end
        table.insert(global["LeosScript"]["RuleList"], arguments[2])
        player.print("Rule was added.")
        return 
    end 

    local ruleNo = tonumber(arguments[2])
    if not ruleNo then player.print("Your rule number argument could not be converted to a number.") return end
    
    if arguments[1] == "remove" then
        if not global["LeosScript"]["RuleList"][ruleNo] then player.print("There is no rule #" .. ruleNo) return end
        table.remove(global["LeosScript"]["RuleList"], ruleNo)
        player.print("Rule #" .. ruleNo .. " was removed.")
        return
    end 

    if not arguments[3] then player.print("you need to supply a rule as 3rd argument.") return end

    if arguments[1] == "replace" then
    if not global["LeosScript"]["RuleList"][ruleNo] then player.print("There is no rule #" .. ruleNo) return end
        global["LeosScript"]["RuleList"][ruleNo] = arguments[3]
        player.print("Rule #" .. ruleNo .. " was replaced.")
        return
    end 

    if arguments[1] == "insert" then
        --[[In array global[][], at position, insert this. (And move the item that was at position down, etc.)]]
        --[[And yes re-ordering the arguments when you add one is a good idea #LUALogic]]
        if not global["LeosScript"]["RuleList"][ruleNo] then player.print("There is no rule #" .. ruleNo) return end
        table.insert(global["LeosScript"]["RuleList"], arguments[2], arguments[3])
        player.print("Rule was inserted as #" .. ruleNo .. ".")
        return
    end

    player.print("Unknown argument(s): \"".. inputString.."\".")
end


local CommandVersionname = "LeosScriptVersion"
local CommandVersionHelp = "Displays the version of LeosScripts that is in use."
local function CommandVersion(event)
    local inputString = event["parameter"]  
    local player = game.players[event["player_index"]]
    if not player then log("Command/gameSpeed: player was nil?") return end 
    player.print("Version: " .. version)
end



--[[Initilisation]]

--[[Initialize on the first tick (version 1.0)]]
--[[This is needen cause on_init does not get called for updates to SoftMods and on_load does not allow writes to global.]]
local ForcedInitOnFirstTickFunctions = {}
local function ForcedInitOnFirstTick(event)
    for key, value in pairs(ForcedInitOnFirstTickFunctions) do
        log("Running FirstTickInit function: " .. key)
        value()
    end
    --[[Change to empty table for release versions?]]
    local ForcedInitOnFirstTickFunctions = function() log("The ForcedInitOnFirstTick event did may not have become unregistered!") end
    script.on_event(defines.events.on_tick, nil)
end

--[[On_Load (version 1.1)]]
local function On_Load()
    log("Leos.On_Load")

    local fakePlayer = {}
    fakePlayer["name"] = "LeosScript.OnLoad"
    
    local bool doFirstTickFix = false
    if not global["LeosScript"] then 
        log("LeosScript is not defined! Attempting to fix on first game tick.")
        ForcedInitOnFirstTickFunctions["No_LeosScript_FullReset"] =
        function()
            global ["LeosScript"] = {} 
            ResetPermissionGroups(true) --[[Will force resetting the permission groups.]]
            ResetSpeedBounds(fakePlayer)
            global["LeosScript"]["walkSpeedModifier"] = 1
        end
        doFirstTickFix = true
    else
        if not global["LeosScript"]["ElevatedPermissionGroups"] then
            log("No ElevatedPermissionGroups defined! Attempting to fix on first game tick.")
            ForcedInitOnFirstTickFunctions["No_ElevatedPermissionGroups"] =
            function()
                ResetPermissionGroups(true) --[[Will force resetting the permission groups.]]
            end
            doFirstTickFix = true
        end
        if not global["LeosScript"]["gameSpeedLowerBound"] or not global["LeosScript"]["gameSpeedLowerBound"] then
            log("No gameSpeedBounds defined! Attempting to fix on first game tick.")
            ForcedInitOnFirstTickFunctions["No_gameSpeedBounds"] =
            function()
                ResetSpeedBounds(fakePlayer)
            end
            doFirstTickFix = true
        end
        if not global["LeosScript"]["walkSpeedModifier"] then
            ForcedInitOnFirstTickFunctions["No_walkSpeedModifier"] =
            function()
                global["LeosScript"]["walkSpeedModifier"] = 1
            end
            doFirstTickFix = true
        end
    end
    if doFirstTickFix then script.on_event(defines.events.on_tick, ForcedInitOnFirstTick) end

    --[[Register commands]]
    commands.remove_command("tp")
    commands.add_command("tp", tpHelp, tp)

    commands.remove_command("tp2me")
    commands.add_command("tp2me", tp2meHelp, tp2me)

    commands.remove_command("tp2spawn")
    commands.add_command("tp2spawn", tp2spawnHelp, tp2spawn)

    commands.remove_command("namethis")
    commands.add_command("namethis", nameOfItemHelp, nameOfItem)

    commands.remove_command("gamespeed")
    commands.add_command("gamespeed", gameSpeedHelp, gameSpeed)

    commands.remove_command(walkSpeedCommand)
    commands.add_command(walkSpeedCommand, walkSpeedHelp, WalkSpeed)

    commands.remove_command("commandpermissions")
    commands.add_command("commandpermissions", CommandPermissionsHelp, CommandPermissions)

    commands.remove_command(CommandRulesName)
    commands.add_command(CommandRulesName, CommandRulesHelp, CommandRules)

    commands.remove_command(CommandVersionname)
    commands.add_command(CommandVersionname, CommandVersionHelp, CommandVersion)
end 

--[[On_Init (version 1.1)]]
local function On_Init()
    log("Leos.On_Init")

    local fakePlayer = {}
    fakePlayer["name"] = "LeosScript.OnLoad"

    ResetPermissionGroups(true) --[[Will force resetting the permission groups.]]
    ResetSpeedBounds(fakePlayer)
    global["LeosScript"]["walkSpeedModifier"] = 1
end


log("Leos.Script:Fullread")

--[[First load]]
script.on_init(
function() 
    On_Init()
end)

--[[Every load, no acess to global!]]
script.on_load(
function() 
    On_Load()
end)

